const LRUMap = require('mnemonist/lru-map');
const Defaults = { max: 500, ttl: 0 };

class LRUCache {
  constructor(max) { this._cache = new LRUMap(max); }

  async put(key, value) { return this._cache.set(key, value); }
  async get(key) { return this._cache.get(key); }

  get capacity() { return this._cache.capacity; }
  get size() { return this._cache.size; }
}

class Cache {
  constructor(options={}) {
    this.options = Object.assign(Defaults, options);

    this.max = this.options.max;
    this.ttl = this.options.ttl;
    this.lru = new LRUCache(this.max);

    this.dbs = (this.options.db || []);
    this.dbs = Array.isArray(this.dbs) ? this.dbs : [];
    this.dbs.unshift(this.lru);
  }

  _expired(item) { return !item || (item.expiry < Date.now()); }
  _valid(item) { return this._expired(item) ? undefined : item; }

  async _set(key, value, dbs=this.dbs) {
    let db;  [db, ...dbs] = dbs;
    await db.put(key, value);
    if (dbs.length) { await this._set(key, value, dbs); }
  }

  async set(key, value, ttl=this.ttl) {
    if (ttl === 0) { return undefined; }
    return this._set(key, {
      value,
      expiry: (new Date()).getTime() + (ttl * 1000)
    });
  }

  async _get(key, dbs=this.dbs) {
    let db, result;
    [db, ...dbs] = dbs;
    result = this._valid(await db.get(key).catch(e => {}));
    if (!result && dbs.length) {
      result = await this._get(key, dbs);
      if(result) { await db.put(key, result); }
    }
    return result;
  }

  async get(key) { return (await this._get(key) || {}).value; }

  async clear(key) { return await this._set(key, undefined); }

  memoize(func, resolver, ttl=this.ttl) {
    resolver = resolver || function() { return Array.from(arguments).join(':'); };

    let memoized = async function() {
      let base = await memoized.base.apply(this, arguments);
      let result = await base.get();
      if (typeof result === 'undefined') {
        result = await base.fetch();
        await base.set(result);
      }
      return result;
    };
    memoized.fresh = async function () {
      let base = await memoized.base.apply(this, arguments);
      let result = await base.fetch();
      await base.set(result);
      return result;
    };
    memoized.peak = async function () {
      let base = await memoized.base.apply(this, arguments);
      let result = await base.get();
      if (typeof result === 'undefined') { result = await base.fetch(); }
      return result;
    };
    memoized.base = function () {
      let self = this,
          args = arguments,
          cache = memoized.cache,
          key = resolver.apply(this, args);

      return {
        async fetch() { return func.apply(self, args); },
        async get() { return cache.get(key); },
        async set(result, _ttl=ttl) {
          _ttl = (typeof _ttl === 'function') ? _ttl : function() { return _ttl; };
          return cache.set(key, result, await Promise.resolve(_ttl.apply(self, args)));
        }
      };
    };
    memoized.cache = this;
    return memoized;
  }
}

module.exports = Cache;
