const cbor = require('cbor');

const Defaults = require('./defaults.json');
const _ = require('./helpers.js');

_.isRecord = (record) => { return (typeof record === 'object'); };
_.isEmptyAddr = (val) => { return val === _.toHex(0,40); };

const Getters = {
  async details() {
    let domain = await this.dwns.registry.domains(this.id);

    return {
      id: this.id,
      updated: domain[0].toString(),
      parent: domain[1],
      label: domain[2],
      fqn: domain[3],
      pointer: domain[4],
      token: domain[5],
      directory: domain[6],
      costInWei: _(domain[7]).convertTo('wei'),
      costInTokens: _(domain[8]).convertTo('tokens'),
      ttl: domain[9].toNumber(),
    };
  },

  async snapshot() {
    let details = await this.details();

    return {
      id: this.id,
      ttl: details.ttl,
      owner: await this.owner(),
      parent: details.parent,
      label: details.label,
      fqn: details.fqn,
      pointer: details.pointer,
      directory: details.directory,
      subdomains: await this.subdomainCount(),
      tld: (details.parent === _.namehash('')),
      cost: {
        tokens: details.costInTokens,
        wei: details.costInWei,
        eth: _(details.costInWei).convertTo('eth')
      },
      token: {
        created: !_.isEmptyAddr(details.token),
        address: details.token,
        supply: await this.totalSupply(),
      },
      record: await this.record()
    };
  },

  async owner() {return this.dwns.token.ownerOf(this.id); },
  async subdomainCount() { return (await this.dwns.registry.subdomainCount(this.id)).toNumber(); },
  async subdomainByIndex(index) { return this.dwns.registry.subdomainByIndex(this.id, index); }
};

const Domain = class Domain {
  constructor(id, dwns) {
    this.id = id;
    this.init(dwns);
  }

  async ready() { return this._ready; }
  async init(dwns) { return this._ready = this._init(dwns); }

  async _init(dwns) {
    this.dwns = await dwns.ready();
    this.ipfs = dwns.ipfs;
    this.cache = dwns.cache;
    this.memoized = {};

    this._cid = await this.ipfs.dag.put(Defaults.record);

    this._get = this.cache.memoize(function(id, key) {
      return Getters[key].apply(this, Array.from(arguments).slice(2));
    }, false, this.ttl);

    this._fresh = this._get.fresh;
    this._peak = this._get.peak;

    await this.ttl();
    return this;
  }

  _memoized(key) { return this._get.base.apply(this, [this.id, key]); };

  async fresh(key) { return _.cloneDeep(await this._fresh(this.id, key)); }

  async get(key) {
    return _.cloneDeep(
      await this._get.apply(this, [this.id].concat(Array.from(arguments)))
    );
  }

  async owner() { return this.get('owner'); }
  async ownedBy(acct) { return acct === (await this.get('owner')); }
  async onlyOwner() {
    let isOwner = await this.ownedBy(this.dwns.account);
    if (!isOwner) { throw new Error('Must Own Domain'); }
  }

  async details(key) {
    let details = await this.get('details');
    return (_.exists(key) ? details[key] : details);
  }

  async ttl() {
    let memoized = this._memoized('details');
    let details = await memoized.get();
    if (typeof details === 'undefined') {
      details = await memoized.fetch();
      await memoized.set(details, (details.ttl || 1));
    }
    return (details.ttl || 1);
  }

  async snapshot(account=this.dwns.account) {
    let snapshot = await this.get('snapshot');
    snapshot.token.balance = _(await this.balance(account)).convertTo('tokens');
    return snapshot;
  }

  async meta() {
    let snapshot = await this.get('snapshot');
    let record = snapshot.record;

    let attrs = [
      {
        key: 'fqn',
        trait_type: 'fqn',
        value: snapshot.fqn,
        type: 'string'
      },
      {
        key: 'subdomains',
        trait_type: 'subdomains',
        value: snapshot.subdomains,
        type: 'integer',
        display_type: 'number'
      }
    ];

    return {
      name: record.name,
      tags: record.meta.tags,
      description: record.description,
      home_url: record.meta.url,
      external_url: record.meta.url,
      image: record.meta.image,
      image_url: record.meta.image,
      background_color: record.meta.color,
      attributes: attrs,
      properties: attrs

    };
  }

  // Tokens
  async token() {
    if(this._token) { return this._token; }

    var addr = await this.details('token');
    if (!_.isEmptyAddr(addr)) {
      return this._token = await this.dwns._contract('RegistrarToken', addr);
    } else { return false; }
  }

  async createToken(name, symbol) {
    await this.dwns.registry.createRegistrarToken(this.id, name, symbol);
    return (await this.fresh('details'))['token'];
  }

  async mint(amount, to=this.dwns.account) {
    console.log('MINTING!!!');
    let token = await this.token();
    await token.mint(to, _(amount.toString()).convertTo('raw', false), '');
    await this.fresh('snapshot');
  }

  async balance(account=this.dwns.account) {
    let token = await this.token();
    if(token) { return token.balanceOf(account); }
    else { return _.toBN(0); }
  }

  async totalSupply() {
    let token = await this.token();
    if(token) { return _(await token.totalSupply()).convertTo('tokens'); }
    else { return '0 tokens'; }
  }

  // Record
  async setRecord(data) {
    let cid = await this.ipfs.dag.put(data);
    await this.dwns.registry.commitPointer(this.id, cid.toBaseEncodedString());
    return (await this.fresh('details'))['pointer'];
  }

  async getRecord(path='/') {
    let cid = await this.details('pointer');
    let result = await this.ipfs.dag.get((cid || this._cid), path);
    return result.value;
  }

  async record(data) {
    if (_.isRecord(data)) { return this.setRecord(data); }
    else { return this.getRecord(data); }
  }

  // Subdomains
  async subdomainCount() { return this.get('subdomainCount'); }
  async subdomainByIndex(index) { return this.get('subdomainByIndex', index); }

  async subdomainAddrs(offset=0, limit=20) {
    let results = [];
    let count = await this.subdomainCount();

    if (offset >= count) { return results; }

    return Promise.all(Array(Math.min(limit, count - offset)).fill().map((x, i) => {
      return this.subdomainByIndex(i + offset);
    }));
  }

  async subdomains(offset, limit) {
    let addrs = await this.subdomainAddrs(offset, limit);
    return Promise.all(addrs.map((addr) => {
      return this.dwns.domain(addr);
    }));
  }

  async register(label, to=this.dwns.account) {
    let snapshot = await this.fresh('snapshot');
    let cost = snapshot.owner === this.dwns.account ? '0' : _(snapshot.cost.wei).convertTo('wei', false);
    await this.dwns.registry.register(to, this.id, label, {value: _(cost)});
    await this.fresh('subdomainByIndex', snapshot.subdomains);
    await this.fresh('subdomainCount');
    return this.fresh('snapshot');
  }

  get cost() {
    let domain = this;
    let wei = async (amount) => {
      if (_(amount).isNil()) { return domain.details('costInWei'); }
      else {
        await domain.dwns.registry.setCostInWei(domain.id, _(amount).convertTo('wei', false));
        await domain.fresh('details');
        return domain.fresh('snapshot');
      }
    };
    wei.tokens = async (amount) => {
      if (_(amount).isNil()) { return domain.details('costInTokens'); }
      else {
        await domain.dwns.registry.setCostInTokens(domain.id, _(amount).convertTo('raw', false));
        await domain.fresh('details');
        return domain.fresh('snapshot');
      }
    };

    return wei;
  }
};

module.exports = Domain;
