const Contract = require('truffle-contract');

const Defaults = require('./defaults.json');
const  _ = require('./helpers');

const Cache = require('./cache.js');
const Domain = require('./domain.js');


const DwNS = class DwNS {
  constructor(options) { if (options) { this._ready = this._init(options); } }

  get _() { return _; }
  get utils() { return _; }

  static async setup(options) { return new DwNS(options).ready(); }

  async ready() {
    if (this._ready) {
      return this._ready;
    } else {
      return this._ready = this._init();
    }
  }

  async init(opts) {
    return this._ready = this._init(opts);
  }

  async _init(opts={}) {
    if (typeof opts === 'function') { opts = opts(); }
    opts = await Promise.resolve(opts);

    let options = this.options = Object.assign({}, this.options, opts);

    this.account = options.from;
    this.ethereum = options.ethereum;
    this.ipfs = options.ipfs;
    this.cache = options.cache || new Cache();

    this.defaults = Object.assign({}, Defaults.contract);
    if (this.account) { this.defaults.from = this.account; }

    this.app = await this._contract('App', options.address);
    this.registry = await this._contract('Registry', this.app.get.call('registry'));
    this.token = await this._contract('OwnershipToken', this.app.get.call('token'));

    return this;
  }

  // Contract
  async _contract(schema, address, defaults={}) {
    let contract = Contract(DwNS.schemas[schema]);
    let addr = await Promise.resolve(address);

    contract.defaults(Object.assign(this.defaults, defaults));
    contract.setProvider(this.ethereum);

    if (addr) {return contract.at(addr); }
    else { return contract.deployed(); }
  }

  // Domains
  async resolve(fqn) { return this.domain(_.namehash(fqn)); }
  async domain(namehash) { return await (new Domain(namehash, this)).ready(); }
  async domains(offset, limit) { return this.domainsOwnedBy(this.account, offset, limit); }

  async ownedDomainCount(owner) { return this.token.balanceOf(owner); }

  async domainOfOwnerByIndex(owner, index) {
    let id = await this.token.tokenOfOwnerByIndex(owner, index);
    return '0x' + id.toString(16).padStart(64, '0');
  }

  async _domainsOwnedBy(owner, offset=0, limit=20) {
    let results = [];
    let count = await this.ownedDomainCount(owner);

    if (offset >= count) { return results; }

    return Promise.all(Array(Math.min(limit, count - offset)).fill().map((x, i) => {
      return this.domainOfOwnerByIndex(owner, i + offset);
    }));
  }

  async domainsOwnedBy(owner, offset, limit) {
    let addrs = await this._domainsOwnedBy(owner, offset, limit);
    return Promise.all(addrs.map((addr) => {
      return this.domain(addr);
    }));
  }
};

DwNS.schemas = {
  App: require('../build/contracts/App.json'),
  Registry: require('../build/contracts/RegistryV0.json'),
  OwnershipToken: require('../build/contracts/OwnershipTokenV0.json'),
  RegistrarToken: require('../build/contracts/RegistrarTokenV0.json')
};

module.exports = DwNS;
