const BigNumber = web3.BigNumber;
const encodeCall = require('zos-lib/lib/helpers/encodeCall').default;

require('chai')
  .use(require('chai-bignumber')(BigNumber))
  .use(require('chai-as-promised'))
  .should();

const Core = artifacts.require('Core');
const Foundry = artifacts.require('FoundryV0');

const Registrar = artifacts.require('RegistrarV0');
const Directory = artifacts.require('DirectoryV0');
const DirectoryToken = artifacts.require('DirectoryTokenV0');
const ZoneToken = artifacts.require('ZoneTokenV0');

contract('Registrar', ([_, owner, investor, benefactor, hacker]) => {
  beforeEach( async () => {
    this.name = "DwNS Token";
    this.symbol = "DWNS";

    this.foundry = await Foundry.deployed();
    this.core = await Core.new(this.foundry.address, this.name, this.symbol);

    var addrs = await this.core.addresses();

    this.directory = await Directory.at(addrs[0]);
    this.registrar = await Registrar.at(addrs[2]);
  });

  describe('properties', () => {
    it( 'should return directory', async () => {
      let result = await this.registrar.directory();
      console.log(result);
      result.should.equal(this.directory.address);
    });
  });

  describe('cost', () => {
    it( 'should allow the owner to set costInTokens', async () => {
      await this.registrar.setCostInTokens(0, 100);
      let result = await this.registrar.costInTokens(0);
      result.should.be.bignumber.equal(100);
    });

    it( 'should allow the owner to set weiPerToken', async () => {
      await this.registrar.setWeiPerToken(0, 1000);
      let result = await this.registrar.weiPerToken(0);
      result.should.be.bignumber.equal(1000);
    });

    it( 'should calculate cost in wei', async () => {
      await this.registrar.setWeiPerToken(0, 30);
      await this.registrar.setCostInTokens(0, 100);
      let result = await this.registrar.cost(0);
      result.should.be.bignumber.equal(3000);
    });
  });

  describe('registrationOpen', () => {
    it('should be closed by default', async () => {
      let result = await this.registrar.registrationOpen(0);
      result.should.be.false;
    });

    it('should allow owner to open registration', async () => {
      await this.registrar.openRegistration(0);
      let result = await this.registrar.registrationOpen(0);
      result.should.be.true;
    });

    it('should allow owner to close registration', async () => {
      await this.registrar.openRegistration(0);
      await this.registrar.closeRegistration(0);
      let result = await this.registrar.registrationOpen(0);
      result.should.be.false;
    });

    it('should allow only the owner to open registration', async () => {
      let result = this.registrar.openRegistration(0, { from: hacker });
      result.should.eventually.be.rejected;
    });

    it('should allow only the owner to close registration', async () => {
      await this.registrar.openRegistration(0);
      let result = this.registrar.closeRegistration(0, { from: hacker });
      result.should.eventually.be.rejected;
    });
  });

  describe('receive tokens', () => {
    beforeEach( async () => {
      let tx = await this.directory.forgeZoneToken(0, 'Root', 'RZT', 1, [_]);
      let addr = tx.logs[3].args.addr;

      this.rootZoneToken = await ZoneToken.at(addr);
    });

    it('should register a zone', async () => {
      await this.rootZoneToken.mint(investor, 250, '');
      await this.registrar.setCostInTokens(0, 10);
      await this.registrar.openRegistration(0);
      await this.rootZoneToken.operatorSend(investor, this.registrar.address, 10, 'test', '');

      let supply = await this.rootZoneToken.totalSupply();
      let owned = await this.directory.zonesOfOwner(investor);
      let zone = await this.directory.zones(1);

      supply.should.be.bignumber.equal(240);
      owned[0].should.be.bignumber.equal(1);
      zone[0].should.be.bignumber.equal(0);
      zone[1].should.equal('test');
      zone[2].should.equal('test');
    });

    it('should add tokens for sale', async () => {
      await this.rootZoneToken.mint(_, 250, '');
      await this.registrar.setCostInTokens(0, 10);
      await this.registrar.openRegistration(0);
      await this.rootZoneToken.operatorSend(_, this.registrar.address, 10, '', '');

      let result = await this.registrar.availableTokens(0);
      result.should.be.bignumber.equal(10);
    });
  });

  describe('register', () => {
    beforeEach( async () => {
      let tx = await this.directory.forgeZoneToken(0, 'Root', 'RZT', 1, [_]);
      let addr = tx.logs[3].args.addr;

      this.rootZoneToken = await ZoneToken.at(addr);
    });

    it('should register a zone if enough eth is sent', async () => {
      await this.rootZoneToken.mint(_, 250, '');
      await this.registrar.setCostInTokens(0, 10);
      await this.registrar.setWeiPerToken(0, 30);
      await this.registrar.openRegistration(0);
      await this.rootZoneToken.operatorSend(_, this.registrar.address, 10, '', '');
      await this.registrar.register(0, 'test', {from: investor, value: 300});

      let supply = await this.rootZoneToken.totalSupply();
      let owned = await this.directory.zonesOfOwner(investor);
      let zone = await this.directory.zones(1);

      supply.should.be.bignumber.equal(240);
      owned[0].should.be.bignumber.equal(1);
      zone[0].should.be.bignumber.equal(0);
      zone[1].should.equal('test');
      zone[2].should.equal('test');
    });
  });


});
