const BigNumber = web3.BigNumber;
const encodeCall = require('zos-lib/lib/helpers/encodeCall').default;

require('chai')
  .use(require('chai-bignumber')(BigNumber))
  .use(require('chai-as-promised'))
  .should();

const ZoneToken = artifacts.require('ZoneTokenV0');

contract('ZoneToken', ([_, owner, tokenHolder, operator, investor, benefactor, hacker]) => {
  beforeEach( async () => {
    this.name = 'DwNS ZoneToken';
    this.symbol = 'DwNS-ZT';
    this.granularity = 2;

    this.token = await ZoneToken.new({ from: _ });

    const callData = encodeCall(
      'initialize',
      ['address', 'string', 'string', 'uint256', 'address[]'],
      [owner, this.name, this.symbol, this.granularity, [owner]]
    );

    await this.token.sendTransaction({data: callData, from: _})
  });

  describe('erc777 properties', () => {
    it('should have a name', async () => {
      const result = await this.token.name();
      result.should.equal(this.name);
    });

    it('should have a symbol', async () => {
      const result = await this.token.symbol();
      result.should.equal(this.symbol);
    });

    it('should have a total supply', async () => {
      const result = await this.token.totalSupply();
      result.should.be.bignumber.equal(0);
    });

    it('should return balanceOf an account', async () => {
      const result = await this.token.balanceOf(investor);
      result.should.be.bignumber.equal(0);
    });

    it('should have a granularity', async () => {
      const result = await this.token.granularity();
      result.should.be.bignumber.equal(2);
    });
  });

  describe('erc777 operator', () => {
    it('should list default operators', async () => {
      const result = await this.token.defaultOperators();
      result.should.eql([owner]);
    });

    it('should consider a tokenHolder to be an operator of their account', async () => {
      const result = await this.token.isOperatorFor(tokenHolder, tokenHolder);
      result.should.be.true;
    });

    it('should consider default operators as operators for tokenHolder', async () => {
      const result = await this.token.isOperatorFor(owner, tokenHolder);
      result.should.be.true;
    });

    it('should consider authorized operators as operators for tokenHolder', async () => {
      await this.token.authorizeOperator(operator, {from: tokenHolder});
      const result = await this.token.isOperatorFor(operator, tokenHolder);
      result.should.be.true;
    });

    it('should allow a tokenHolder to revoke an authorized operator', async () => {
      await this.token.authorizeOperator(operator, {from: tokenHolder});
      await this.token.revokeOperator(operator, {from: tokenHolder});
      const result = await this.token.isOperatorFor(operator, tokenHolder);
      result.should.be.false;
    });

    it('should allow a tokenHolder to revoke a default operator', async () => {
      await this.token.revokeOperator(owner, {from: tokenHolder});
      const result = await this.token.isOperatorFor(owner, tokenHolder);
      result.should.be.false;
    });

    it('should not allow unauthorized operators', async () => {
      const result = await this.token.isOperatorFor(hacker, tokenHolder);
      result.should.be.false;
    });
  });

  describe('minting', () => {
    it('should allow the owner to mint tokens', async () => {
      await this.token.mint(tokenHolder, 100, "", {from: owner});
      const result = await this.token.totalSupply();
      result.should.be.bignumber.equal(100);
    });

    it('should assign the minted tokens to the benefactor', async () => {
      await this.token.mint(tokenHolder, 100, "", {from: owner});
      const result = await this.token.balanceOf(tokenHolder);
      result.should.be.bignumber.equal(100);
    });


    it('should reject if minting is not multiple of granularity', async () => {
      const result = this.token.mint(tokenHolder, 3, "", {from: owner});
      result.should.eventually.be.rejected;
    });

    it('should reject if minting is finished', async () => {
      await this.token.finishMinting({from: owner});
      const result = this.token.mint(tokenHolder, 100, "", {from: owner});
      result.should.eventually.be.rejected;
    });
  });

  describe('send', () => {
    it('should send tokens', async () => {
      await this.token.mint(tokenHolder, 1000, "", {from: owner});
      const callData = encodeCall(
        'send', ['address', 'uint256', 'bytes'],
        [benefactor, 100, ""]
      );

      await this.token.sendTransaction({data: callData, from: tokenHolder});
      const result = await this.token.balanceOf(benefactor);
      result.should.be.bignumber.equal(100);
    });
  });

});
